/*
 * Copyright (c) 2015-2016 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <synthpod_lv2.h>

#include <lv2/lv2plug.in/ns/ext/atom/atom.h>
#include <lv2/lv2plug.in/ns/ext/atom/forge.h>
#include <lv2/lv2plug.in/ns/ext/atom/util.h>
#include <lv2/lv2plug.in/ns/ext/midi/midi.h>

typedef struct _plughandle_t plughandle_t;

struct _plughandle_t {
	const float *audio_in [4];
	const float *xfade;
	float *audio_out [4];
};

static LV2_Handle
instantiate(const LV2_Descriptor* descriptor, double rate,
	const char *bundle_path, const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
		return NULL;
	mlock(handle, sizeof(plughandle_t));

	return handle;
}

static void
connect_port(LV2_Handle instance, uint32_t port, void *data)
{
	plughandle_t *handle = instance;

	switch(port)
	{
		case 0:
			handle->audio_in[0] = (const float *)data;
			break;
		case 1:
			handle->audio_in[1] = (const float *)data;
			break;
		case 2:
			handle->audio_in[2] = (const float *)data;
			break;
		case 3:
			handle->audio_in[3] = (const float *)data;
			break;
		case 4:
			handle->xfade = (const float *)data;
			break;
		case 5:
			handle->audio_out[0] = (float *)data;
			break;
		case 6:
			handle->audio_out[1] = (float *)data;
			break;
		case 7:
			handle->audio_out[2] = (float *)data;
			break;
		case 8:
			handle->audio_out[3] = (float *)data;
			break;
		default:
			break;
	}
}

static void
activate(LV2_Handle instance)
{
	//plughandle_t *handle = instance;
	
	// nothing
}

__realtime static void
run(LV2_Handle instance, uint32_t nsamples)
{
	plughandle_t *handle = instance;

	const float b = *handle->xfade;
	const float a = 1.f - b;

	for(uint32_t i=0; i<nsamples; i++)
	{
		const float x = handle->audio_in[0][i];
		const float v = handle->audio_in[1][i];
		const float y = handle->audio_in[2][i];
		const float w = handle->audio_in[3][i];

		handle->audio_out[0][i] = a*x + b*y;
		handle->audio_out[1][i] = a*v + b*w;
		handle->audio_out[2][i] = b*x + a*y;
		handle->audio_out[3][i] = b*v + a*w;
	}
}

static void
deactivate(LV2_Handle instance)
{
	//plughandle_t *handle = instance;

	// nothing
}

static void
cleanup(LV2_Handle instance)
{
	plughandle_t *handle = instance;

	munlock(handle, sizeof(plughandle_t));
	free(handle);
}

static const void*
extension_data(const char* uri)
{
	return NULL;
}

const LV2_Descriptor synthpod_audioxfadestereo = {
	.URI						= SYNTHPOD_AUDIOXFADESTEREO_URI,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= activate,
	.run						= run,
	.deactivate			= deactivate,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};
